Fix saving MMX registers.

BIG NOTE: the whole save/restore cycle must be done under a spinlock.

Index: kernel/zaptel-base.c
===================================================================
--- a/kernel/zaptel-base.c	(revision 4539)
+++ b/kernel/zaptel-base.c	(working copy)
@@ -337,7 +337,48 @@ static int zt_chan_ioctl(struct inode *i
 /* XXX kernel_fpu_begin() is NOT exported properly (in 2.4), so we have to make
        a local version.  Somebody fix this! XXX */
 
-#ifndef LINUX26
+#if (defined(CONFIG_X86) && !defined(CONFIG_X86_64)) || defined(CONFIG_I386)
+struct fpu_save_buf {
+	unsigned long cr0;
+	unsigned long fpu_buf[128];
+};
+
+static DEFINE_PER_CPU(struct fpu_save_buf, fpu_buf);
+
+/** zt_kernel_fpu_begin() - Save floating point registers
+ *
+ * This function is similar to kernel_fpu_begin() . However it is
+ * designed to work in an interrupt context. Restoring must be done with
+ * zt_kernel_fpu_end().
+ *
+ * Furthermore, the whole code between the call to
+ * zt_kernel_fpu_begin() and zt_kernel_fpu_end() must reside
+ * inside a spinlock. Otherwise the context might be restored to the
+ * wrong process.
+ *
+ * Current implementation is x86/ia32-specific and will not even build on 
+ * x86_64)
+ * */
+static inline void zt_kernel_fpu_begin(void)
+{
+	struct fpu_save_buf *buf = &__get_cpu_var(fpu_buf);
+	__asm__ __volatile__ ("movl %%cr0,%0; clts" : "=r" (buf->cr0));
+	__asm__ __volatile__ ("fnsave %0" : "=m" (buf->fpu_buf));
+}
+
+/** zt_kernel_fpu_end() - restore floating point context
+ *
+ * Must be used with context saved by zt_kernel_fpu_begin(). See its
+ * documentation for further information.
+ */
+static inline void zt_kernel_fpu_end(void)
+{
+	struct fpu_save_buf *buf = &__get_cpu_var(fpu_buf);
+	__asm__ __volatile__ ("frstor %0" : "=m" (buf->fpu_buf));
+	__asm__ __volatile__ ("movl %0,%%cr0" : : "r" (buf->cr0));
+}
+
+#elif not defiend(LINUX26)
 static inline void __save_init_fpu( struct task_struct *tsk )
 {
 	if ( cpu_has_fxsr ) {
@@ -359,8 +375,14 @@ static inline void zt_kernel_fpu_begin(v
 	}
 	clts();
 }
-#else
+#define zt_kernel_fpu_end   kernel_fpu_end
+#else /* Linux 2.6, but we haven't fixed FP context saving/restoring yet */
+/* Very strange things can happen when the context is not properly
+ * restored. OTOH, some people do report success with this. Hence we 
+ * so far just issue a warning */
+#warning CONFIG_ZAPTEL_MMX may behave randomely on this platform
 #define zt_kernel_fpu_begin kernel_fpu_begin
+#define zt_kernel_fpu_end   kernel_fpu_end
 #endif /* LINUX26 */
 #endif	
 
@@ -6364,7 +6386,7 @@ static inline void __zt_ec_chunk(struct 
 #endif /* defined(ZT_EC_ARRAY_UPDATE) */
 		}
 #if defined(CONFIG_ZAPTEL_MMX) || defined(ECHO_CAN_FP)
-		kernel_fpu_end();
+		zt_kernel_fpu_end();
 #endif		
 	}
 	spin_unlock_irqrestore(&ss->lock, flags);
@@ -7266,7 +7288,7 @@ static void __zt_transmit_chunk(struct z
 #endif
 		__zt_process_getaudio_chunk(chan, buf);
 #ifdef CONFIG_ZAPTEL_MMX
-		kernel_fpu_end();
+		zt_kernel_fpu_end();
 #endif
 	}
 }
@@ -7351,7 +7373,7 @@ static void __zt_receive_chunk(struct zt
 #endif
 		__zt_process_putaudio_chunk(chan, buf);
 #ifdef CONFIG_ZAPTEL_MMX
-		kernel_fpu_end();
+		zt_kernel_fpu_end();
 #endif
 	}
 	__zt_putbuf_chunk(chan, buf);
@@ -7569,7 +7591,7 @@ int zt_receive(struct zt_span *span)
 				}
 			}
 #ifdef CONFIG_ZAPTEL_MMX
-			kernel_fpu_end();
+			zt_kernel_fpu_end();
 #endif			
 		}
 		/* do all the pseudo/conferenced channel transmits (putbuf's) */
